/*
 * @OSF_COPYRIGHT@
 * COPYRIGHT NOTICE
 * Copyright (c) 1996 Open Software Foundation, Inc.
 * ALL RIGHTS RESERVED (DCE).  See the file named COPYRIGHT.DCE for
 * the full copyright text.
 * 
 */
/*
 * @HP_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: dcpAppInit.c,v $
 * Revision 1.1.2.2  1996/03/09  20:43:50  marty
 * 	Add OSF copyright
 * 	[1996/03/09  20:23:49  marty]
 *
 * Revision 1.1.2.1  1995/12/08  21:20:36  root
 * 	Submit OSF/DCE 1.2.1
 * 
 * 	HP revision /main/HPDCE02/5  1995/05/03  19:51 UTC  jrr
 * 	Fix init of global login list mutex (CHFts15101).
 * 
 * 	HP revision /main/HPDCE02/jrr_dcecp_chfts15101/1  1995/04/26  23:05 UTC  jrr
 * 	Move init of global login list mutex to dcpAppInit().
 * 
 * 	HP revision /main/HPDCE02/4  1995/02/21  16:16 UTC  truitt
 * 	Merge convenience variable work.
 * 
 * 	HP revision /main/HPDCE02/truitt_dcecp_chfts14/1  1995/02/21  16:15 UTC  truitt
 * 	CHFts14390: Define two variables which will track
 * 	when the _o and _e variables are set.
 * 
 * 	HP revision /main/HPDCE02/3  1994/10/21  21:31 UTC  pulsinelli
 * 	Fix dynamic linkage for cdsbrowser.
 * 
 * 	HP revision /main/HPDCE02/2  1994/10/21  20:59 UTC  pulsinelli
 * 	Remove "-c" command line option.
 * 
 * 	HP revision /main/pulsinelli_dcecp_sl/1  1994/10/19  19:18 UTC  pulsinelli
 * 	Initial creation.
 * 	[1995/12/08  17:55:29  root]
 * 
 * $EndLog$
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <locale.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>

#include <dce/tcl.h>
#include <dce/dce.h>
#include <dce/editline.h>
#include <dce/dce_msg.h>
#include <dce/dcesvcmsg.h>
#include <dce/dcedcpmsg.h>

#include <dcecp.h>
#include <util_login.h>


/* Global Variables */
Tcl_Interp *attr_interp;
Tcl_Interp *dir_interp;
boolean32 check_visible_pw = FALSE;
boolean32 was_e_var_set = FALSE;
boolean32 was_o_var_set = FALSE;


/* shouldn't these be generated by sams? */
static dce_svc_subcomp_t dcp_svc_table[5];
static dce_svc_handle_t  dcp_svc_handle;


/* External Declarations */
extern void init_cds_attribute_table(Tcl_Interp *);
extern void init_rgy_attribute_table(Tcl_Interp *);


/* Foward Refrences */
static void init_dcp_svc(void);


/* 
 * DCECP Application specific initialization
 */

int dcpAppInit(Tcl_Interp *interp)
{
    char          **argv;
    char           *cds_attr_eval_file = "attr_eval.tcl";
    char           *cds_dir_ops_file   = "dir_ops.dcecp";
    char            temp_attr_eval_file[256];
    char            command[1024];
    boolean32       command_flag = FALSE;
    int             result = TCL_OK;


    /* Initialize the msg table */
    init_dcp_svc();
    
    /* Initialize dcecp commands */
    dcp_initInterp(interp);
    

    /*
     * Create/Initialize the Attribute interpreter
     */
    if (result == TCL_OK) {
        attr_interp = Tcl_CreateInterp();
        init_cds_attribute_table(interp);
        init_rgy_attribute_table(interp);

        sprintf(temp_attr_eval_file, "%s/%s/%s", dcelocal_path, DCECP_SCRIPT_DIR, cds_attr_eval_file);
        sprintf(command, "if [file exists %s] {source %s} else "
                         "{error \"Attribute evaluation script attr_eval.tcl does not exist in appropriate directory. \"}",
                         temp_attr_eval_file, temp_attr_eval_file);

        if ((result = Tcl_Eval(attr_interp, command) != TCL_OK)) {
            printf("%s\n", attr_interp->result);
        }
    }


    /*
     * Set up scripts for directory -tree operations
     */
    if (result == TCL_OK) {
        dir_interp = Tcl_CreateInterp();
        dcp_initInterp(dir_interp);

        sprintf(command, "if [file exists $dcecp_library/%s] {source $dcecp_library/%s} else "
                         "{error \"Directory operations script dir_ops.dcecp does not exist in appropriate directory. \"}",
                         cds_dir_ops_file, cds_dir_ops_file);

        if ((result = Tcl_Eval(dir_interp, command) != TCL_OK)) {
            printf("%s\n", dir_interp->result);
        }
    }


    /* Initialize the mutex for the global_context_list. */
    if (pthread_mutex_init(&login_list, pthread_mutexattr_default) != 0) {
        DCP_SET_RESULT_CODE(dcp_s_mutex_init_fail);
        return TCL_ERROR;
    }
    
    /* Check to see if an initial login context exists */    
    if (result == TCL_OK) {
        if (result = check_initial_auth(interp) == TCL_ERROR) {
            DCP_SET_RESULT_CODE(dcp_s_initial_context_fail);
        }
    }


    return result;
}


static void init_dcp_svc(void)
{
    error_status_t	        st;
    
    /* Load message table; ignore errors. */
    dce_msg_define_msg_table(dcp_msg_table,
                             sizeof dcp_msg_table / sizeof dcp_msg_table[0],
                             &st);

    if (st != msg_s_ok) {
        (void)fprintf(stderr, "Cannot register dcp msg table, error 0x%lx\n", st);
        exit(1);
    }    
    
    dcp_svc_handle = dce_svc_register(dcp_svc_table, (idl_char *)"dcp", &st);

    if (dcp_svc_handle == NULL) {
        (void)fprintf(stderr, "Cannot register svc table, error 0x%lx\n", st);
        exit(1);
    }

}
